import os
import shutil
import tempfile
from zipfile import ZipFile, ZIP_STORED, ZipInfo


class UpdateableZipFile(ZipFile):
    """
    Add delete (via remove_file) and update (via writestr and write methods)
    To enable update features use UpdateableZipFile with the 'with statement',
    Upon  __exit__ (if updates were applied) a new zip file will override the exiting one with the updates

    Credits: nosklo and Or Weis; https://stackoverflow.com/a/35435548
    """

    class DeleteMarker(object):
        pass

    def __init__(self, file, mode="r", compression=ZIP_STORED, allowZip64=False):
        # Init base
        super(UpdateableZipFile, self).__init__(file, mode=mode,
                                                compression=compression,
                                                allowZip64=allowZip64)
        # track file to override in zip
        self._replace = {}
        # Whether the with statement was called
        self._allow_updates = False

    def writestr(self, zinfo_or_arcname, data, compress_type=None):
        if isinstance(zinfo_or_arcname, ZipInfo):
            name = zinfo_or_arcname.filename
        else:
            name = zinfo_or_arcname
        # If the file exits, and needs to be overridden,
        # mark the entry, and create a temp-file for it
        # we allow this only if the with statement is used
        if self._allow_updates and name in self.namelist():
            temp_file = self._replace[name] = self._replace.get(name,
                                                                tempfile.TemporaryFile())
            temp_file.write(data.encode('utf-8'))
        # Otherwise just act normally
        else:
            super(UpdateableZipFile, self).writestr(zinfo_or_arcname,
                                                    data.encode('utf-8'), compress_type=compress_type)

    def write(self, filename, arcname=None, compress_type=None):
        arcname = arcname or filename
        # If the file exits, and needs to be overridden,
        # mark the entry, and create a temp-file for it
        # we allow this only if the with statement is used
        if self._allow_updates and arcname in self.namelist():
            temp_file = self._replace[arcname] = self._replace.get(arcname,
                                                                   tempfile.TemporaryFile())
            with open(filename, "rb") as source:
                shutil.copyfileobj(source, temp_file)
        # Otherwise just act normally
        else:
            super(UpdateableZipFile, self).write(filename, 
                                                 arcname=arcname, compress_type=compress_type)

    def __enter__(self):
        # Allow updates
        self._allow_updates = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # call base to close zip file, organically
        try:
            super(UpdateableZipFile, self).__exit__(exc_type, exc_val, exc_tb)
            if len(self._replace) > 0:
                self._rebuild_zip()
        finally:
            # In case rebuild zip failed,
            # be sure to still release all the temp files
            self._close_all_temp_files()
            self._allow_updates = False

    def _close_all_temp_files(self):
        for temp_file in self._replace.values():
            if hasattr(temp_file, 'close'):
                temp_file.close()

    def remove_file(self, path):
        self._replace[path] = self.DeleteMarker()

    def _rebuild_zip(self):
        tempdir = tempfile.mkdtemp()
        try:
            temp_zip_path = os.path.join(tempdir, 'new.zip')
            with ZipFile(self.filename, 'r') as zip_read:
                # Create new zip with assigned properties
                with ZipFile(temp_zip_path, 'w', compression=self.compression,
                             allowZip64=self._allowZip64) as zip_write:
                    for item in zip_read.infolist():
                        # Check if the file should be replaced / or deleted
                        replacement = self._replace.get(item.filename, None)
                        # If marked for deletion, do not copy file to new zipfile
                        if isinstance(replacement, self.DeleteMarker):
                            del self._replace[item.filename]
                            continue
                        # If marked for replacement, copy temp_file, instead of old file
                        elif replacement is not None:
                            del self._replace[item.filename]
                            # Write replacement to archive,
                            # and then close it (deleting the temp file)
                            replacement.seek(0)
                            data = replacement.read()
                            replacement.close()
                        else:
                            data = zip_read.read(item.filename)
                        zip_write.writestr(item, data)
            # Override the archive with the updated one
            shutil.move(temp_zip_path, self.filename)
        finally:
            shutil.rmtree(tempdir)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ghost_pwa']

package_data = \
{'': ['*'], 'ghost_pwa': ['icons/*']}

install_requires = \
['requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'ghost-pwa',
    'version': '2.0.3',
    'description': '',
    'long_description': '# Ghost Theme PWA\n\nThis theme enables you to "Install" your site as a Progressive Web App.\n\nRequires either Chrome or Firefox to install the app.\n\nBrowsing works better with Chrome. Editing works better with Firefox.\n\n## Theme ZIP\n\nIf you just have a ZIP archive, run:\n\n```\n$ make_pwa <archive file>\n```\n\nExample:\n\n```\n$ make_pwa Casper-3.0.7.zip\n```\n\nThe output is:\n\n```\n$ ls\nCasper-3.0.7-PWA.zip\n```\n\n## Verifying\n\nTest your PWA here: https://www.seochecker.it/pwa-tester-online\n\n## Use online\n\nTo use this online, you can open a notebook in Google Colab and apply the steps\ndescribed above.\n\nDownload a ZIP file with your current theme. Next, upload it to Colab. Finally,\nrun the script on the ZIP file, and download the result. Voila! You can now\nupload the PWA ZIP file to your Ghost installation.\n\nUsage example in a notebook:\n\nhttps://colab.research.google.com/drive/1cPrGzrS15Nz_7OhIt8Y8FfOD_wqM0H67\n\n## Development\n\n### Test\n\n`pytest --doctest-modules`\n',
    'author': 'Patryk Kocielnik',
    'author_email': 'patryk@kocielnik.pl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from os import path, remove
from pathlib import Path
import requests
import shutil
from subprocess import run
from tempfile import TemporaryDirectory

try:
    from updateable_zip_file import UpdateableZipFile as ZipFile, ZipInfo
except ModuleNotFoundError:
    from ghost_pwa.updateable_zip_file import (
        UpdateableZipFile as ZipFile,
        ZipInfo
    )

header_insert = """
    {{!-- PWA --}}
    <link rel="manifest" href="/manifest.webmanifest">
    <meta name="theme-color" content="#0075FF">
    <link rel="apple-touch-icon" href="/assets/icons/ghost-4.png">
"""

footer_insert = """
{{!-- Service Worker for PWA --}}
<script>
if ('serviceWorker' in navigator) {
    window.addEventListener('load', () => {
    navigator.serviceWorker.register('/sw.js');
    });
}
</script>
"""

MODULE_DIR = path.dirname(path.realpath(__file__))

def get_position_for_insert(tag, text):
    start_position = 0
    end_position = text.index("{{" + tag + "}}")
    input_position = text.rfind("\n\n", start_position, end_position) + 1
    return input_position

def insert_before_tag(tag, insert, original):
    input_position = get_position_for_insert(tag, original)

    return "".join([
        original[:input_position],
        insert,
        original[input_position:]
    ])

def insert_pwa_scripts(original_markup):
    with_header = insert_before_tag(
        "ghost_head", header_insert, original_markup
    )
    with_header_and_footer = insert_before_tag(
        "ghost_foot", footer_insert, with_header
    )
    return with_header_and_footer

def get_file(url, out_file):
    r = requests.get(url, allow_redirects=True)
    open(out_file, 'wb').write(r.content)

def dir_exists(directory):
    return path.isdir(directory)

def unpack_example():
    """
    >>> unpack_example()
    """
    with TemporaryDirectory() as temp_dir:
        get_file(
            "https://codeload.github.com/TryGhost/Casper/zip/refs/tags/3.1.3",
            f"{temp_dir}/Casper-3.1.3.zip"
        )
        unpack(f"{temp_dir}/Casper-3.1.3.zip", target_dir=temp_dir)
        assert dir_exists(f"{temp_dir}/Casper-3.1.3")

    assert dir_exists(temp_dir) is False

def unpack(archive, target_dir="."):
    shutil.unpack_archive(archive, target_dir)

def get_root_dir(archive_file):
    """
    Returns the root directory of the theme ZIP file.

    Examples:

    >>> with ExampleArchiveWithOneSubdir("test.zip") as test_archive:
    ...     get_root_dir(test_archive)
    'Casper-3.1.3'

    >>> with EmptyExampleArchive("test.zip") as test_archive:
    ...     get_root_dir(test_archive)
    '.'

    >>> with ExampleArchiveWithTwoSubdirs("test.zip") as test_archive:
    ...     get_root_dir(test_archive)
    '.'
    """
    with ZipFile(archive_file) as zip_file:
        paths = [
            file[:-1] for file in zip_file.namelist()
            if file.count("/") == 1 and file[-1] == '/'
        ]

        return paths[0] if len(paths) == 1 else '.'

def get_new_file_path(original_file_path):
    """
    >>> path = Path("Casper.zip")
    >>> new_path_obj = get_new_file_path(path)
    >>> str(new_path_obj)
    'Casper-PWA.zip'
    """
    return Path(original_file_path.stem + "-PWA" + original_file_path.suffix)

def get_processed_zip(archive_file):
    original_file = Path(archive_file)
    new_file = get_new_file_path(original_file)

    shutil.copy(original_file, new_file)

    with ZipFile(new_file) as out_zip:
        root_dir = get_root_dir(new_file)
        default_hbs_path = f"{root_dir}/default.hbs"
        assert default_hbs_path in out_zip.namelist()

        default_hbs = out_zip.read(default_hbs_path).decode()

    with_header_and_footer = insert_pwa_scripts(default_hbs)

    with ZipFile(new_file, 'a') as out_zip:
        out_zip.writestr(default_hbs_path, with_header_and_footer)

        out_zip.write(
            f"{MODULE_DIR}/manifest.webmanifest",
            f"{root_dir}/manifest.webmanifest"
        )

        out_zip.write(
            f"{MODULE_DIR}/sw.js",
            f"{root_dir}/sw.js"
        )

        out_zip.write(
            f"{MODULE_DIR}/icons/ghost-2.png",
            f"{root_dir}/assets/icons/ghost-2.png"
        )
        out_zip.write(
            f"{MODULE_DIR}/icons/ghost-4.png",
            f"{root_dir}/assets/icons/ghost-4.png"
        )

    return new_file

class ExampleArchive:
    def __init__(self, name):
        self.name = name

    def __enter__(self):
        self._create_archive(self.name)
        return self.name

    def __exit__(self, type, value, traceback):
        remove(self.name)

    def _create_archive(self, name):
        with ZipFile(name, 'w') as out_zip:
            pass

def mkdir(zip_file, folder):
    zfi = ZipInfo(folder + '/')
    zip_file.writestr(zfi, '')

class EmptyExampleArchive(ExampleArchive):
    pass

class ExampleArchiveWithOneSubdir(ExampleArchive):
    def _create_archive(self, name):
        with ZipFile(name, 'w') as out_zip:
            mkdir(out_zip, "Casper-3.1.3")

class ExampleArchiveWithTwoSubdirs(ExampleArchive):
    def _create_archive(self, name):
        with ZipFile(name, 'w') as out_zip:
            mkdir(out_zip, "assets")
            mkdir(out_zip, "doc")

def make_pwa(file_name):
    new_file = get_processed_zip(file_name)
    return new_file

def example_make_casper_pwa():
    """
    Example:
    >>> example_make_casper_pwa()
    PosixPath('Casper-3.1.3-PWA.zip')

    Clean up:
    >>> remove("Casper-3.1.3.zip")
    >>> remove("Casper-3.1.3-PWA.zip")
    """
    get_file(
        "https://codeload.github.com/TryGhost/Casper/zip/refs/tags/3.1.3",
        "Casper-3.1.3.zip"
    )
    return make_pwa("Casper-3.1.3.zip")
